import wrapt

from ...log import logger
from ...sdk import sdk

import grpc


def instrument():
    def instrument_grpc_callable(wrapped, instance, args, kwargs):
        try:
            target = instance._channel.target()
            method = instance._method

            if isinstance(method, bytes):
                method = method.decode("utf-8")

            if isinstance(target, bytes):
                target = target.decode("utf-8")

            url = f"grpc://{target}{method}"

        except Exception as e:
            logger.debug("Could not instrument grpc call, error: '{}'", e)
            return wrapped(*args, **kwargs)

        with sdk.trace_outgoing_web_request(url, "POST") as tracer:
            logger.debug("Tracing GRPC, url: '{}'".format(url))
            ret = wrapped(*args, **kwargs)
            try:
                rpc_state = ret[0]
                sdk.add_custom_request_attribute("Status code", "{}".format(rpc_state.code))
                if rpc_state.code == grpc.StatusCode.OK:
                    tracer.set_status_code(200)
                else:
                    tracer.set_status_code(500)
            finally:
                return ret

    @wrapt.patch_function_wrapper("grpc._channel", "_UnaryUnaryMultiCallable._blocking")
    def unary_call_dynatrace(wrapped, instance, args, kwargs):
        return instrument_grpc_callable(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("grpc._channel", "_UnaryStreamMultiCallable.__call__")
    def unary_stream_call_dynatrace(wrapped, instance, args, kwargs):
        return instrument_grpc_callable(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("grpc._channel", "_SingleThreadedUnaryStreamMultiCallable.__call__")
    def single_unary_call_dynatrace(wrapped, instance, args, kwargs):
        return instrument_grpc_callable(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("grpc._channel", "_StreamUnaryMultiCallable._blocking")
    def stream_unary_blocking_dynatrace(wrapped, instance, args, kwargs):
        return instrument_grpc_callable(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("grpc._channel", "_StreamStreamMultiCallable.__call__")
    def stream_call_dynatrace(wrapped, instance, args, kwargs):
        return instrument_grpc_callable(wrapped, instance, args, kwargs)



from django.db import connections

from ...log import logger
from ...sdk import sdk
from ..utils import normalize_vendor
from ..dbapi import TracedCursor

import oneagent

import socket

CURSOR_ATTR = "_dynatrace_cursor"
ALL_CONNS_ATTR = "_dynatrace_connections_all"


def instrument_db():
    if hasattr(connections, ALL_CONNS_ATTR):
        return

    setattr(connections, ALL_CONNS_ATTR, connections.all)

    def all_connections(self, *args, **kwargs):
        conns = getattr(self, ALL_CONNS_ATTR)()
        for conn in conns:
            instrument_conn(conn)
        return conns

    connections.all = all_connections.__get__(connections, type(connections))


def instrument_conn(conn):
    if hasattr(conn, CURSOR_ATTR):
        return

    setattr(conn, CURSOR_ATTR, conn.cursor)

    def cursor():
        alias = getattr(conn, "alias", "default")
        vendor = normalize_vendor(getattr(conn, "vendor", "db"))
        logger.debug("Instrumenting the DB {} {}".format(alias, vendor))

        db_technology = conn.settings_dict.get("ENGINE", "SQL").split(".")[-1]
        db_name = conn.settings_dict.get("NAME", "Unknow")
        db_host = conn.settings_dict.get("HOST", None)
        db_port = conn.settings_dict.get("PORT", None)

        channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.OTHER, None)
        if db_host and db_port:
            try:
                socket.inet_pton(socket.AF_INET6, db_host)
                db_host = "[{}]".format(db_host)
            except Exception:
                pass
            channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(db_host, db_port))

        db_info = sdk.create_database_info(str(db_name), str(db_technology), channel)
        return TracedCursor(conn._dynatrace_cursor(), db_info)

    conn.cursor = cursor



import fastapi
from fastapi.middleware import Middleware
from .middleware import DynatraceASGIMiddleware
import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    @wrapt.patch_function_wrapper("fastapi.applications", "FastAPI.__init__")
    def init_dynatrace(wrapped, instance, args, kwargs):
        middlewares = kwargs.pop("middleware", [])
        middlewares.insert(0, Middleware(DynatraceASGIMiddleware))
        kwargs.update({"middleware": middlewares})
        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("fastapi", "routing.run_endpoint_function")
    async def session_init_dynatrace(wrapped, instance, args, kwargs):
        try:
            name = kwargs.get("dependant").call.__name__
        except Exception:
            name = "run_endpoint_function"

        logger.debug("Tracing fastapi.{}".format(name))
        with sdk.trace_custom_service(name, "FastAPI"):
            return await wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("fastapi", "routing.serialize_response")
    async def serialize_response_dynatrace(wrapped, instance, args, kwargs):
        logger.debug("Tracing fastapi.routing.serialize_response")
        with sdk.trace_custom_service("serialize_response", "FastAPI"):
            return await wrapped(*args, **kwargs)



import oneagent
from django.conf import settings
from django.utils.module_loading import import_string
from django_stomp import execution
from ...log import logger
from ...sdk import sdk
from .utils import get_messaging_type_by_queue_name

def instrument_consumer():
    def wrapped_import_string(dotted_path):
        callback_function = import_string(dotted_path)

        def instrumented_callback(payload):
            try:
                headers = payload.headers
                host, port = settings.STOMP_SERVER_HOST, settings.STOMP_SERVER_PORT
                destination = headers.get("tshoot-destination")
            except Exception as e:
                logger.warn("autodynatrace - Could not trace Consumer(import_string): {}".format(e))
                return callback_function(payload)

            tag = None
            if headers is not None:
                tag = headers.get(oneagent.common.DYNATRACE_MESSAGE_PROPERTY_NAME, None)

            messaging_system = sdk.create_messaging_system_info(
                oneagent.common.MessagingVendor.RABBIT_MQ,
                destination,
                get_messaging_type_by_queue_name(destination),
                oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(host, port)),
            )

            with messaging_system:
                with sdk.trace_incoming_message_receive(messaging_system):
                    with sdk.trace_incoming_message_process(messaging_system, str_tag=tag) as process_message:
                        process_message.set_correlation_id(headers.get("correlation-id"))
                        process_message.set_vendor_message_id(headers.get("message-id"))
                        logger.debug(
                            f"autodynatrace - Tracing Incoming RabbitMQ host={host}, port={port},"
                            f" routing_key={destination}, tag={tag}, headers={headers}"
                        )
                        return callback_function(payload)

        return instrumented_callback

    setattr(execution, "import_string", wrapped_import_string)



import logging

logger = logging.getLogger("autodynatrace")


def init(level=logging.INFO):
    ch = logging.StreamHandler()
    f = logging.Formatter(
        "%(asctime)s: %(process)d %(levelname)s %(name)s - %(funcName)s: %(message)s"
    )
    ch.setFormatter(f)
    logger.addHandler(ch)
    logger.setLevel(level)



from mock import patch

from autodynatrace.wrappers.django.utils import get_host, get_request_uri, get_app_name


@patch("django.http.HttpRequest")
def test_get_host(django_request_mock):
    django_request_mock.get_host = lambda: "myhost:80"
    assert get_host(django_request_mock) == "myhost:80"

    django_request_mock.get_host = lambda: None
    django_request_mock.META = {"HTTP_HOST": "myhost:80"}
    assert get_host(django_request_mock) == "myhost:80"

    django_request_mock.META = {"SERVER_NAME": "myhost", "SERVER_PORT": 80}
    assert get_host(django_request_mock) == "myhost:80"

    django_request_mock.META = {}
    assert get_host(django_request_mock) == "unknown"

    django_request_mock = None
    assert get_host(django_request_mock) == "unknown"


@patch("django.http.HttpRequest")
def test_get_request_uri(django_request_mock):
    django_request_mock.get_host = lambda: "myhost:80"
    django_request_mock.scheme = "http"
    django_request_mock.path = "/path"

    assert get_request_uri(django_request_mock) == "http://myhost:80/path"


@patch("django.http.HttpRequest")
def test_get_app_name(django_request_mock):
    django_request_mock.path = "/path"
    django_request_mock.META = {"SERVER_NAME": "myhost", "SERVER_PORT": 80}
    assert get_app_name(django_request_mock) == "Django (myhost:80)"



import socket

import oneagent
import sqlalchemy
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.engine.base import Connection
import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    @event.listens_for(Engine, "before_cursor_execute", named=True)
    def dynatrace_before_cursor_execute(**kw):

        try:
            conn = kw["conn"]
            context = kw["context"]

            statement = kw.get("statement", "")
            db_technology = conn.engine.name
            db_name = conn.engine.url.database

            if not db_name:
                # The connection string might not have a db name
                db_name = "default"

            db_host = conn.engine.url.host
            db_port = conn.engine.url.port

            channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.OTHER, None)
            if db_host is not None and db_port is not None:
                try:
                    socket.inet_pton(socket.AF_INET6, db_host)
                    db_host = "[{}]".format(db_host)
                except Exception:
                    pass
                channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(db_host, db_port))

            db_info = sdk.create_database_info(db_name, db_technology, channel)
            tracer = sdk.trace_sql_database_request(db_info, "{}".format(statement))
            logger.debug("Tracing SQLAlchemy: '{}, {}@{}:{}".format(statement, db_name, db_host, db_port))
            tracer.start()
            context.dynatrace_tracer = tracer

        except Exception as e:
            logger.debug("Error instrumenting sqlalchemy: {}".format(e))

    @event.listens_for(Engine, "after_cursor_execute", named=True)
    def dynatrace_after_cursor_execute(**kw):

        try:
            context = kw["context"]

            if context is not None and hasattr(context, "dynatrace_tracer"):
                tracer = context.dynatrace_tracer
                if tracer is not None:
                    # TODO Check if I get stats about query
                    tracer.end()

        except Exception as e:
            logger.debug("Error instrumenting sqlalchemy: {}".format(e))

    @event.listens_for(Engine, "handle_error", named=True)
    def dynatrace_handle_error(**kw):

        try:
            context = kw["exception_context"]

            if context is not None and hasattr(context, "dynatrace_tracer"):
                tracer = context.dynatrace_tracer
                if tracer is not None:
                    logger.debug("Marking purepath as failed")
                    tracer.mark_failed_exc()
                    tracer.end()

        except Exception as e:
            logger.debug("Error instrumenting sqlalchemy: {}".format(e))

    @wrapt.patch_function_wrapper("sqlalchemy.orm.session", "Session.__init__")
    def session_init_dynatrace(wrapped, instance, args, kwargs):
        logger.debug("Tracing sqlalchemy.Session.init")
        with sdk.trace_custom_service("Session.init", "sqlalchemy"):
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("sqlalchemy.orm.session", "Session.begin")
    def session_init_dynatrace(wrapped, instance, args, kwargs):
        logger.debug("Tracing sqlalchemy.Session.begin")
        with sdk.trace_custom_service("Session.begin", "sqlalchemy"):
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("sqlalchemy.orm.session", "Session.connection")
    def session_init_dynatrace(wrapped, instance, args, kwargs):
        logger.debug("Tracing sqlalchemy.Session.connection")
        with sdk.trace_custom_service("Session.connection", "sqlalchemy"):
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("sqlalchemy.orm.session", "Session.close")
    def session_init_dynatrace(wrapped, instance, args, kwargs):
        logger.debug("Tracing sqlalchemy.Session.close")
        with sdk.trace_custom_service("Session.close", "sqlalchemy"):
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("sqlalchemy.orm.session", "Session.query")
    def session_init_dynatrace(wrapped, instance, args, kwargs):
        logger.debug("Tracing sqlalchemy.Session.query")
        with sdk.trace_custom_service("Session.query", "sqlalchemy"):
            return wrapped(*args, **kwargs)



from pymongo import monitoring
import oneagent

from ...log import logger
from ...sdk import sdk


def instrument():
    class DynatraceMongoListener(monitoring.CommandListener):
        def __init__(self):
            self.vendor = "MongoDB"
            self._tracer_dict = {}

        def started(self, event):

            try:
                db_name = event.database_name
                collection_name = event.command.get(event.command_name)
                operation = "{} at {}".format(event.command_name, collection_name)
                host, port = event.connection_id
                channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(host, port))

                db_info = sdk.create_database_info(db_name, self.vendor, channel)
                tracer = sdk.trace_sql_database_request(db_info, "{}".format(operation))
                self._tracer_dict[_get_tracer_dict_key(event)] = tracer
                tracer.start()
                logger.debug("Tracing Mongo call: {}({})@{}:{}".format(operation, db_name, host, port))

            except Exception as e:
                logger.debug("Error instrumenting MongoDB: {}".format(e))

        def succeeded(self, event):
            self.end(False, event)

        def failed(self, event):
            self.end(True, event, message="{}".format(event.failure))

        def end(self, failed, event, message=""):
            tracer = self._tracer_dict.get(_get_tracer_dict_key(event))
            if tracer is not None:
                if event and not isinstance(event, monitoring.CommandSucceededEvent):
                    logger.debug("Got bad pymongo event: {}".format(event))
                    tracer.mark_failed("MongoDB Command", message)

                tracer.end()
                self._tracer_dict.pop(_get_tracer_dict_key(event))

    monitoring.register(DynatraceMongoListener())

    def _get_tracer_dict_key(event):
        if event.connection_id is not None:
            return event.request_id, event.connection_id
        return event.request_id



import os

from ...sdk import sdk
from ...log import logger


class DynatraceASGIMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):

        # If this is not an http request, do nothing and return
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        request_headers = scope.get("headers", [])
        headers = {}
        for key, value in request_headers:
            key = key.decode() if isinstance(key, bytes) else key
            value = value.decode() if isinstance(value, bytes) else value
            headers[key.lower()] = value

        dt_tag = headers.get("x-dynatrace")

        host = "{}:{}".format(scope["server"][0], scope["server"][1])
        url = "{}://{}{}?{}".format(scope["scheme"], host, scope["path"], scope["query_string"].decode())
        method = scope.get("method", "GET")
        app = scope.get("app")
        app_name = app.title if app is not None else "FastAPI"
        root = scope.get("root_path", "/")

        virtual_host = os.environ.get("AUTODYNATRACE_VIRTUAL_HOST", "{}".format(host))
        app_name = os.environ.get("AUTODYNATRACE_APPLICATION_ID", "{}".format(app_name))
        context_root = os.environ.get("AUTODYNATRACE_CONTEXT_ROOT", root or "/")

        with sdk.create_web_application_info(virtual_host, app_name, context_root) as web_app_info:
            with sdk.trace_incoming_web_request(web_app_info, url, method, headers=headers, str_tag=dt_tag) as tracer:
                logger.debug("dynatrace - asgi - Tracing url: '{}' with tag {}".format(url, dt_tag))

                async def wrapped_send(message):
                    if message.get("type") == "http.response.start" and "status" in message:
                        tracer.set_status_code(message["status"])

                    if "headers" in message:
                        response_headers = {}
                        for k, v in message["headers"]:
                            k = k.decode() if isinstance(k, bytes) else k
                            v = v.decode() if isinstance(v, bytes) else v
                            response_headers[k.lower()] = v

                        tracer.add_response_headers(response_headers)

                    return await send(message)

                return await self.app(scope, receive, wrapped_send)



import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    @wrapt.patch_function_wrapper("ruxit.api.base_plugin", "BasePlugin._query_internal")
    def query_dynatrace(wrapped, instance, args, kwargs):
        plugin_name = type(instance).__name__
        with sdk.trace_custom_service("query", plugin_name):
            logger.debug("Custom tracing - {}: {}".format(plugin_name, "query"))
            return wrapped(*args, **kwargs)



import psycopg2
import wrapt
import oneagent
import functools

from ...log import logger
from ...sdk import sdk


def instrument():
    def parse_dsn(dsn):
        return {c.split("=")[0]: c.split("=")[1] for c in dsn.split() if "=" in c}

    class DynatraceCursor(psycopg2.extensions.cursor):
        def __init__(self, *args, **kwargs):
            self._dynatrace_db_info = kwargs.pop("dynatrace_db_info", None)
            super(DynatraceCursor, self).__init__(*args, **kwargs)

        def execute(self, query, vars=None):
            if hasattr(self, "_dynatrace_db_info") and self._dynatrace_db_info is not None:
                with sdk.trace_sql_database_request(self._dynatrace_db_info, "{}".format(query)) as tracer:
                    try:
                        logger.debug("Tracing psycopg2 query: '{}'".format(query))
                        return super(DynatraceCursor, self).execute(query, vars)
                    finally:
                        tracer.set_rows_returned(self.rowcount)

            return super(DynatraceCursor, self).execute(query, vars)

    class DynatraceConnection(psycopg2.extensions.connection):
        def __init__(self, *args, **kwargs):
            super(DynatraceConnection, self).__init__(*args, **kwargs)

            dsn = parse_dsn(self.dsn)
            self._dynatrace_db_info = sdk.create_database_info(
                dsn.get("dbname", "unknown"),
                oneagent.sdk.DatabaseVendor.POSTGRESQL,
                oneagent.sdk.Channel(
                    oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(dsn.get("host", "localhost"), dsn.get("port", 5432))
                ),
            )

            self._dynatrace_cursor = functools.partial(DynatraceCursor, dynatrace_db_info=self._dynatrace_db_info)

        def cursor(self, *args, **kwargs):
            kwargs.setdefault("cursor_factory", self._dynatrace_cursor)
            return super(DynatraceConnection, self).cursor(*args, **kwargs)

    @wrapt.patch_function_wrapper("psycopg2", "connect")
    def dynatrace_connect(wrapped, instance, args, kwargs):
        kwargs.setdefault("connection_factory", DynatraceConnection)
        return wrapped(*args, **kwargs)



from typing import Optional

import oneagent


def get_messaging_type_by_queue_name(queue_name: Optional[str] = None) -> int:
    """Helper function to get queue type by name. If 'VirtualTopic' exists in the destination name
    this queue is a TOPIC type.
    By default the type of queue is QUEUE.
    """
    if queue_name and "VirtualTopic" in queue_name:
        return oneagent.common.MessagingDestinationType.TOPIC

    return oneagent.common.MessagingDestinationType.QUEUE



import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    @wrapt.patch_function_wrapper("paramiko.client", "SSHClient.connect")
    def paramiko_connect(wrapped, instance, args, kwargs):
        try:
            host = args[0]
        except IndexError:
            host = kwargs.get("hostname","")
        port = kwargs.get("port", 22)
        url = "ssh://{}:{}".format(host, port)

        with sdk.trace_outgoing_web_request(url, "POST") as tracer:
            logger.debug("Tracing paramiko SSHClient.connect, url: '{}'".format(url))
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("paramiko.client", "SSHClient.exec_command")
    def paramiko_exec_command(wrapped, instance, args, kwargs):
        cmd = args[0]
        with sdk.trace_custom_service("exec_command", "Paramiko"):
            logger.debug("Tracing paramiko SSHClient.exec_command, cmd: '{}'".format(cmd))
            sdk.add_custom_request_attribute("Command", cmd)
            return wrapped(*args, **kwargs)



from six import text_type

VALUE_PLACEHOLDER = "?"
VALUE_MAX_LEN = 100
VALUE_TOO_LONG_MARK = "..."
CMD_MAX_LEN = 1000


def format_command_args(args):

    length = 0
    out = []
    for arg in args:
        try:
            cmd = text_type(arg)

            if len(cmd) > VALUE_MAX_LEN:
                cmd = cmd[:VALUE_MAX_LEN] + VALUE_TOO_LONG_MARK

            if length + len(cmd) > CMD_MAX_LEN:
                prefix = cmd[: CMD_MAX_LEN - length]
                out.append("%s%s" % (prefix, VALUE_TOO_LONG_MARK))
                break

            out.append(cmd)
            length += len(cmd)
        except Exception:
            out.append(VALUE_PLACEHOLDER)
            break

    return " ".join(out)



import starlette
import asyncio

import wrapt
from ...log import logger
from ...sdk import sdk


def instrument():
    @wrapt.patch_function_wrapper("starlette.responses", "Response.__init__")
    def starlette_parsing_dynatrace(wrapped, instance, args, kwargs):

        # FastAPI creates a empty response at the beginning of a request, ignore it
        if kwargs.get("status_code") is None:
            return wrapped(*args, **kwargs)

        logger.debug("Tracing starlette.Response.__init__")
        with sdk.trace_custom_service("Response.render", "starlette"):
            return wrapped(*args, **kwargs)

    # TODO - Add asyncio support
    # @wrapt.patch_function_wrapper("asyncio", "BaseEventLoop.create_task")
    # def starlette_parsing_dynatrace(wrapped, instance, args, kwargs):
    #     logger.debug("Tracing asyncio.BaseEventLoop.create_task")
    #     with sdk.trace_custom_service("BaseEventLoop.create_task", "asyncio"):
    #         return wrapped(*args, **kwargs)

    # loop = asyncio.get_event_loop()
    # if not isinstance(loop.create_task, wrapt.ObjectProxy):
    #
    #     @wrapt.patch_function_wrapper(loop, "create_task")
    #     def starlette_parsing_dynatrace(wrapped, instance, args, kwargs):
    #         logger.debug("Tracing asyncio.loop.create_task")
    #         with sdk.trace_custom_service("loop.create_task", "asyncio"):
    #             return wrapped(*args, **kwargs)



import aiohttp
import wrapt

from oneagent.common import DYNATRACE_HTTP_HEADER_NAME
from ...log import logger
from ...sdk import sdk


def instrument():
    @wrapt.patch_function_wrapper("aiohttp.client", "ClientSession._request")
    async def dynatrace_request(wrapped, instance, args, kwargs):

        method = args[0]
        url = args[1]
        headers = dict(kwargs.get("headers", {}))

        with sdk.trace_outgoing_web_request(url, method, headers) as tracer:
            tag = tracer.outgoing_dynatrace_string_tag.decode()
            logger.debug("dynatrace - tracing {} '{}' with tag '{}'".format(method, url, tag))
            headers.update({DYNATRACE_HTTP_HEADER_NAME: tag})
            kwargs["headers"] = headers

            response = await wrapped(*args, **kwargs)

            tracer.set_status_code(response.status)
            tracer.add_response_headers(dict(response.headers))

            return response



import oneagent
import wrapt

from ...log import logger
from ...sdk import sdk
from django.conf import settings
from django_outbox_pattern.producers import Producer
from django_outbox_pattern import headers


def instrument_publisher():
    def on_send_message(wrapped, instance, args, kwargs):
        try:
            host, port = settings.DJANGO_OUTBOX_PATTERN["DEFAULT_STOMP_HOST_AND_PORTS"][0]
            destination = kwargs.get("destination")
            headers = kwargs.get("headers", {})

        except Exception as e:
            logger.warn("autodynatrace - Could not trace Publisher.send: {}".format(e))
            return wrapped(**kwargs)

        messaging_system = sdk.create_messaging_system_info(
            oneagent.common.MessagingVendor.RABBIT_MQ,
            destination,
            oneagent.common.MessagingDestinationType.QUEUE,
            oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(host, port)),
        )

        tag = headers.get(oneagent.common.DYNATRACE_MESSAGE_PROPERTY_NAME, None)
        with messaging_system:
            with sdk.trace_outgoing_message(messaging_system) as outgoing_message:
                outgoing_message.set_correlation_id(headers.get("correlation-id"))
                if not tag:
                    tag = outgoing_message.outgoing_dynatrace_string_tag.decode("utf-8")
                    headers[oneagent.common.DYNATRACE_MESSAGE_PROPERTY_NAME] = tag
                logger.debug(
                    f"autodynatrace - Tracing Outgoing RabbitMQ host={host}, port={port}, routing_key={destination}, "
                    f"tag={tag}, headers={headers}"
                )
                return wrapped(**kwargs)

    wrapt.wrap_function_wrapper(Producer, "_send_with_retry", on_send_message)

    original_get_message_headers = headers.get_message_headers

    def on_get_message_headers(message):
        headers = original_get_message_headers(message)
        try:
            host, port = settings.DJANGO_OUTBOX_PATTERN["DEFAULT_STOMP_HOST_AND_PORTS"][0]
            destination = headers.get("destination")
        except Exception as e:
            logger.warn("autodynatrace - Could not trace Publisher.send: {}".format(e))
            return headers

        messaging_system = sdk.create_messaging_system_info(
            oneagent.common.MessagingVendor.RABBIT_MQ,
            destination,
            oneagent.common.MessagingDestinationType.QUEUE,
            oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(host, port)),
        )

        tag = headers.get(oneagent.common.DYNATRACE_MESSAGE_PROPERTY_NAME, None)
        with sdk.trace_custom_service('publish', 'django_outbox_pattern'):
            with messaging_system:
                with sdk.trace_outgoing_message(messaging_system) as outgoing_message:
                    outgoing_message.set_correlation_id(headers.get("correlation-id"))
                    if not tag:
                        tag = outgoing_message.outgoing_dynatrace_string_tag.decode("utf-8")
                        headers[oneagent.common.DYNATRACE_MESSAGE_PROPERTY_NAME] = tag
                    logger.debug(
                        f"autodynatrace - Tracing Outgoing RabbitMQ host={host}, port={port}, routing_key={destination}, "
                        f"tag={tag}, headers={headers}"
                    )
                    return headers

    setattr(headers, "get_message_headers", on_get_message_headers)





import six
import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    def dynatrace_trace(wrapped, instance, args, kwargs):
        method = wrapped.__name__
        with sdk.trace_custom_service(method, "Subprocess"):
            logger.debug("Tracing subprocess.{}".format(method))
            sdk.add_custom_request_attribute("args", str(args))
            return wrapped(*args, **kwargs)

    if six.PY3:

        @wrapt.patch_function_wrapper("subprocess", "run")
        def dynatrace_run(wrapped, instance, args, kwargs):
            return dynatrace_trace(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("subprocess", "call")
    def dynatrace_call(wrapped, instance, args, kwargs):
        return dynatrace_trace(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("subprocess", "check_call")
    def dynatrace_check_call(wrapped, instance, args, kwargs):
        return dynatrace_trace(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("subprocess", "check_output")
    def dynatrace_check_output(wrapped, instance, args, kwargs):
        return dynatrace_trace(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("subprocess", "Popen")
    def dynatrace_run(wrapped, instance, args, kwargs):
        return dynatrace_trace(wrapped, instance, args, kwargs)



import importlib
import logging
import threading
import os
import sys

from wrapt.importer import when_imported

from .log import init as log_init, logger
from .sdk import init as sdk_init

__version__ = 1.082

dt_custom_prop = os.environ.get("DT_CUSTOM_PROP")
if dt_custom_prop:
    os.environ["DT_CUSTOM_PROP"] = "{} Autodynatrace={}".format(dt_custom_prop, __version__)
else:
    os.environ["DT_CUSTOM_PROP"] = "Autodynatrace={}".format(__version__)


log_level_name = os.environ.get("AUTODYNATRACE_LOG_LEVEL", "WARNING")
log_init(logging.getLevelName(log_level_name))

forkable = os.environ.get("AUTODYNATRACE_FORKABLE", "False").strip().lower() in ("true", "1")
sdk_init(forkable=forkable)

from .wrappers.custom import dynatrace_custom_tracer as trace

_LOCK = threading.Lock()
_INSTRUMENTED_LIBS = set()
_INSTRUMENT_LIBS_LAZY = set()

INSTRUMENT_LIBS = {
    "flask": True,
    "celery": True,
    "pymongo": True,
    "sqlalchemy": True,
    "django": True,
    "redis": True,
    "pika": True,
    "cx_Oracle": True,
    "grpc": True,
    "ruxit": True,
    "confluent_kafka": True,
    "pysnmp": True,
    "concurrent": False,
    "urllib": True,
    "suds": True,
    "subprocess": True,
    "paramiko": True,
    "psycopg2": True,
    "tornado": True,
    "fastapi": True,
    "starlette": True,
    "aiohttp": True,
    "bottle": True,
    "django_stomp": True,
    "django_outbox_pattern": True,
}


def will_instrument(lib_name, default):
    # Check if the user has chose to forcefully instrument (or not instrument) this lib
    lib_environment_variable_name = "AUTODYNATRACE_INSTRUMENT_{}".format(lib_name.upper())

    if lib_environment_variable_name not in os.environ:
        # If the environment variable was not set, return the default instrumentation setting
        return default

    return os.environ.get(lib_environment_variable_name, "True").strip().lower() in ("true", "1")


def load(_):
    pass


def instrument_all():
    libs = INSTRUMENT_LIBS.copy()
    instrument(libs)


def _on_import_wrapper(lib):
    def on_import(hook):

        path = "autodynatrace.wrappers.%s" % lib
        try:
            logger.debug("Instrumenting imported lib '{}'".format(lib))
            imported_module = importlib.import_module(path)
            imported_module.instrument()
        except Exception:
            logger.debug("Could not instrument {}".format(lib), exc_info=True)

    return on_import


def instrument(instrument_libs):

    for lib_name, default in instrument_libs.items():

        if will_instrument(lib_name, default):

            if lib_name in sys.modules:
                instrument_lib(lib_name)
                _INSTRUMENTED_LIBS.add(lib_name)

            else:
                when_imported(lib_name)(_on_import_wrapper(lib_name))
                _INSTRUMENT_LIBS_LAZY.add(lib_name)

    patched_libs = get_already_instrumented()
    lazy_libs = get_will_instrument()
    logger.info(
        "Instrumented {}/{} libraries ({}). Will instrument when imported: ({})".format(
            len(patched_libs), len(instrument_libs), ", ".join(patched_libs), ", ".join(lazy_libs)
        )
    )


def instrument_lib(lib):
    try:
        logger.debug("Attempting to instrument %s", lib)
        return _instrument_lib(lib)
    except Exception:
        logger.debug("Failed to instrument %s", lib, exc_info=True)
        return False


def _instrument_lib(lib):

    path = "autodynatrace.wrappers.%s" % lib
    with _LOCK:
        if lib in _INSTRUMENTED_LIBS and lib:
            logger.debug("Skipping (already instrumented): {}".format(path))
            return False

        imported_module = importlib.import_module(path)
        imported_module.instrument()
        _INSTRUMENTED_LIBS.add(lib)
        return True


def get_already_instrumented():
    with _LOCK:
        return sorted(_INSTRUMENTED_LIBS)


def get_will_instrument():
    with _LOCK:
        return sorted(_INSTRUMENT_LIBS_LAZY)


instrument_all()



import wrapt

from ...log import logger
from ...sdk import sdk

from pysnmp.proto.rfc1905 import GetRequestPDU, GetBulkRequestPDU


def instrument():
    @wrapt.patch_function_wrapper("pysnmp.proto.rfc3412", "MsgAndPduDispatcher.sendPdu")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        try:
            host, port = args[2]
            request = args[10]

            oids = []
            if isinstance(request, (GetBulkRequestPDU, GetRequestPDU)):
                oids = [str(vb["name"]) for vb in request["variable-bindings"]]

            url = "snmp://{}:{}?query={}".format(host, port, ",".join(oids))

        except Exception as e:
            logger.debug("Could not instrument sendPdu: {}".format(e))
            return wrapped(*args, **kwargs)

        with sdk.trace_outgoing_web_request(url, "POST") as tracer:
            logger.debug("Tracing pysnmp sendPdu, url: '{}'".format(url))
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.proto.rfc3412", "MsgAndPduDispatcher.receiveMessage")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("receiveMessage", "SNMP"):
            logger.debug("Tracing pysnmp receiveMessage")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.proto.rfc3412", "MsgAndPduDispatcher.returnResponsePdu")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("returnResponsePdu", "SNMP"):
            logger.debug("Tracing pysnmp returnResponsePdu")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.entity.rfc3413.cmdgen", "CommandGenerator.processResponsePdu")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("processResponsePdu", "SNMP"):
            logger.debug("Tracing pysnmp processResponsePdu")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.entity.rfc3413.cmdgen", "BulkCommandGeneratorSingleRun.sendVarBinds")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("sendVarBinds", "SNMP"):
            logger.debug("Tracing pysnmp sendVarBinds")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.hlapi.asyncore.cmdgen", "bulkCmd")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("cmdgen.bulkCmd", "SNMP"):
            logger.debug("Tracing pysnmp cmdgen.bulkCmd")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.hlapi.asyncore", "bulkCmd")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("asyncore.bulkCmd", "SNMP"):
            logger.debug("Tracing pysnmp asyncore.bulkCmd")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.hlapi.asyncore", "getCmd")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("asyncore.getCmd", "SNMP"):
            logger.debug("Tracing pysnmp asyncore.getCmd")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.hlapi", "bulkCmd")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("sync.bulkCmd", "SNMP"):
            logger.debug("Tracing pysnmp sync.bulkCmd")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.entity.engine", "SnmpEngine.__init__")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("SnmpEngine()", "SNMP"):
            logger.debug("Tracing pysnmp SnmpEngine.__init__")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.hlapi.lcd", "CommandGeneratorLcdConfigurator.configure")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("configure", "SNMP"):
            logger.debug("Tracing pysnmp CommandGeneratorLcdConfigurator.configure")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.hlapi.varbinds", "CommandGeneratorVarBinds.makeVarBinds")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("makeVarBinds", "SNMP"):
            logger.debug("Tracing pysnmp CommandGeneratorVarBinds.makeVarBinds")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pysnmp.hlapi.asyncore.cmdgen", "getCmd")
    def send_pdu_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("cmdgen.getCmd", "SNMP"):
            logger.debug("Tracing pysnmp cmdgen.getCmd")
            return wrapped(*args, **kwargs)



import wrapt

from ...log import logger
from ...sdk import sdk
from oneagent.common import MessagingDestinationType
from oneagent.sdk import Channel, ChannelType
from threading import Thread, local

import confluent_kafka

threadlocal = local()
threadlocal.tracer = None

class Producer(confluent_kafka.Producer):
    pass


class Consumer(confluent_kafka.Consumer):
    pass


def instrument():
    confluent_kafka.Consumer = Consumer
    confluent_kafka.Producer = Producer

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.confluent_kafka.wrapper", "Producer.__init__")
    def custom_producer_init(wrapped, instance, args, kwargs):

        servers = args[0].get("bootstrap.servers", "unknown-host")
        setattr(instance, "dt_servers", servers)

        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.confluent_kafka.wrapper", "Consumer.__init__")
    def custom_consumer_init(wrapped, instance, args, kwargs):

        servers = args[0].get("bootstrap.servers", "unknown-host")
        setattr(instance, "dt_servers", servers)

        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.confluent_kafka.wrapper", "Producer.produce")
    def custom_produce(wrapped, instance, args, kwargs):

        servers = getattr(instance, "dt_servers", "unknown-host")
        topic = args[0] if args else kwargs.get("topic", "unknown-topic")

        msi_handle = sdk.create_messaging_system_info(
            "Kafka",
            topic,
            MessagingDestinationType.TOPIC,
            Channel(ChannelType.TCP_IP, servers),
        )

        with msi_handle:
            with sdk.trace_outgoing_message(msi_handle) as tracer:
                tag = tracer.outgoing_dynatrace_string_tag
                logger.debug("kafka-producer Injecting message with header '{}'".format(tag))
                headers = kwargs.get("headers", {})
                headers.update({"dtdTraceTagInfo": tag})
                kwargs["headers"] = headers
                return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.confluent_kafka.wrapper", "Consumer.poll")
    def custom_poll(wrapped, instance, args, kwargs):
        if threadlocal.tracer is not None:
           threadlocal.tracer.end()
           threadlocal.tracer = None
        message = wrapped(*args, **kwargs)
        if message is not None:
            try:
                servers = getattr(instance, "dt_servers", "unknown-host")
                msi_handle = sdk.create_messaging_system_info(
                    "Kafka",
                    message.topic(),
                    MessagingDestinationType.TOPIC,
                    Channel(ChannelType.TCP_IP, servers),
                )
                with msi_handle:
                    tag = None
                    headers = message.headers()
                    if headers is not None:
                        for header in headers:
                            if header[0] == "dtdTraceTagInfo":
                                tag = header[1]
                    tracer = sdk.trace_incoming_message_process(msi_handle, str_tag=tag)
                    tracer.start()
                    threadlocal.tracer = tracer
                    logger.debug("kafka-consumer: Received message with tag {}".format(tag))
                    return message
            except Exception:
                logger.debug("Could not trace Consumer.poll", exc_info=True)
                return message
        return message



import django
from django.conf import settings
import os


from ...log import logger
from ...sdk import sdk
from ..utils import func_name

from .utils import get_request_uri, get_host, get_app_name

try:
    from django.utils.deprecation import MiddlewareMixin

    MiddlewareClass = MiddlewareMixin
except ImportError:
    MiddlewareClass = object


class DynatraceMiddleware(MiddlewareClass):
    def process_request(self, request):
        try:
            url = get_request_uri(request)
            logger.debug("Tracing request {}".format(url))
            host = get_host(request)
            method = request.method
            app_name = get_app_name(request)

            headers = {}
            dt_header = request.META.get("HTTP_X_DYNATRACE", None)
            if dt_header is not None:
                headers = {"x-dynatrace": dt_header}

            if os.environ.get("AUTODYNATRACE_CAPTURE_HEADERS", False):
                headers.update(getattr(request, "headers", {}))

            virtual_host = os.environ.get("AUTODYNATRACE_VIRTUAL_HOST", host)
            context_root = os.environ.get("AUTODYNATRACE_CONTEXT_ROOT", "/")

            with sdk.create_web_application_info(virtual_host, app_name, context_root) as web_app_info:
                tracer = sdk.trace_incoming_web_request(web_app_info, url, method, headers=headers)
                _set_req_tracer(request, tracer)
                tracer.start()

        except Exception:
            logger.error("Error tracing request", exc_info=True)

    def process_view(self, request, view_func, *args, **kwargs):
        name = func_name(view_func)
        logger.debug("Starting view tracer {}".format(name))
        tracer = sdk.trace_custom_service(name, "Django Views")
        _add_child_tracer(request, tracer)
        tracer.start()

    def process_response(self, request, response):
        try:

            # First, end all children
            for child in _get_child_tracers(request):
                if child:
                    child.end()

            tracer = _get_req_tracer(request)
            if tracer:
                tracer.set_status_code(response.status_code)
                tracer.end()

        except Exception:
            logger.debug("Error processing response", exc_info=True)
        finally:
            return response


def _get_req_tracer(request):
    return getattr(request, "_dynatrace_tracer", None)


def _set_req_tracer(request, tracer):
    return setattr(request, "_dynatrace_tracer", tracer)


def _add_child_tracer(request, tracer):
    tracers = _get_child_tracers(request)
    tracers.append(tracer)
    setattr(request, "_dynatrace_child_tracers", tracers)


def _get_child_tracers(request):
    return getattr(request, "_dynatrace_child_tracers", [])


def get_middleware_insertion_point():
    middleware = getattr(settings, "MIDDLEWARE", None)
    if middleware is not None and django.VERSION >= (1, 10):
        return "MIDDLEWARE", middleware
    return "MIDDLEWARE_CLASSES", getattr(settings, "MIDDLEWARE_CLASSES", None)


def insert_dynatrace_middleware():
    middleware_attribute, middleware = get_middleware_insertion_point()
    if middleware is not None:
        setattr(
            settings,
            middleware_attribute,
            type(middleware)(("autodynatrace.wrappers.django.middlewares.DynatraceMiddleware",)) + middleware,
        )



import sys
import os

import autodynatrace
import autodynatrace.wrappers.custom.wrapper as custom_wrapper


@autodynatrace.trace
def module_function():
    return 1


def another_function():
    pass


class MyClass:
    def class_method(self):
        pass

    @staticmethod
    def static_method(self):
        pass


@autodynatrace.trace("Service")
def decorated_service_only():
    return 1


@autodynatrace.trace(method="method")
def decorated_method_only():
    return 1


@autodynatrace.trace("Service", "method")
def decorated_service_and_method():
    return 1

def test_custom_service_name():
    my_class = MyClass()

    os.environ.pop("AUTODYNATRACE_CUSTOM_SERVICE_USE_FQN", None)
    os.environ.pop("AUTODYNATRACE_CUSTOM_SERVICE_NAME", None)
    assert custom_wrapper.generate_service_name(module_function) == "tests.test_custom"
    assert custom_wrapper.generate_service_name(my_class.class_method) == "MyClass"
    assert custom_wrapper.generate_service_name(classmethod(my_class.class_method)) == "MyClass"
    assert custom_wrapper.generate_service_name(module_function) == custom_wrapper.generate_service_name(another_function)

    if sys.version_info[0] == 2:
        assert custom_wrapper.generate_service_name(my_class.static_method) == "tests.test_custom"
        assert custom_wrapper.generate_service_name(MyClass.static_method) == "tests.test_custom"
    else:
        assert custom_wrapper.generate_service_name(my_class.static_method) == "MyClass"
        assert custom_wrapper.generate_service_name(MyClass.static_method) == "MyClass"

    os.environ["AUTODYNATRACE_CUSTOM_SERVICE_NAME"] = "CustomServiceName"
    assert custom_wrapper.generate_service_name(module_function) == "CustomServiceName"


def test_custom_service_name_fqn_true():
    my_class = MyClass()

    os.environ["AUTODYNATRACE_CUSTOM_SERVICE_USE_FQN"] = "TRUE"
    os.environ.pop("AUTODYNATRACE_CUSTOM_SERVICE_NAME", None)
    assert custom_wrapper.generate_service_name(module_function) == "tests.test_custom"
    assert custom_wrapper.generate_service_name(my_class.class_method) == "tests.test_custom.MyClass"
    assert custom_wrapper.generate_service_name(classmethod(my_class.class_method)) == "tests.test_custom.MyClass"
    assert custom_wrapper.generate_service_name(module_function) == custom_wrapper.generate_service_name(another_function)

    if sys.version_info[0] == 2:
        assert custom_wrapper.generate_service_name(my_class.static_method) == "tests.test_custom"
        assert custom_wrapper.generate_service_name(MyClass.static_method) == "tests.test_custom"
    else:
        assert custom_wrapper.generate_service_name(my_class.static_method) == "tests.test_custom.MyClass"
        assert custom_wrapper.generate_service_name(MyClass.static_method) == "tests.test_custom.MyClass"

    os.environ["AUTODYNATRACE_CUSTOM_SERVICE_NAME"] = "CustomServiceName"
    assert custom_wrapper.generate_service_name(module_function) == "CustomServiceName"


def test_custom_method_name_fqn_false():
    my_class = MyClass()

    os.environ.pop("AUTODYNATRACE_CUSTOM_SERVICE_USE_FQN", None)
    os.environ.pop("AUTODYNATRACE_CUSTOM_SERVICE_NAME", None)
    assert custom_wrapper.generate_method_name(module_function) == "module_function"
    assert custom_wrapper.generate_method_name(my_class.class_method) == "class_method"
    assert custom_wrapper.generate_method_name(classmethod(my_class.class_method)) == "class_method"
    assert custom_wrapper.generate_method_name(another_function) == "another_function"

    assert custom_wrapper.generate_method_name(my_class.static_method) == "static_method"
    assert custom_wrapper.generate_method_name(MyClass.static_method) == "static_method"

    os.environ["AUTODYNATRACE_CUSTOM_SERVICE_NAME"] = "CustomServiceName"
    assert custom_wrapper.generate_method_name(module_function) == "tests.test_custom.module_function"
    assert custom_wrapper.generate_method_name(my_class.class_method) == "MyClass.class_method"
    assert custom_wrapper.generate_method_name(classmethod(my_class.class_method)) == "MyClass.class_method"
    assert custom_wrapper.generate_method_name(another_function) == "tests.test_custom.another_function"

    if sys.version_info[0] == 2:
        assert custom_wrapper.generate_method_name(my_class.static_method) == "tests.test_custom.static_method"
        assert custom_wrapper.generate_method_name(MyClass.static_method) == "tests.test_custom.static_method"
    else:
        assert custom_wrapper.generate_method_name(my_class.static_method) == "MyClass.static_method"
        assert custom_wrapper.generate_method_name(MyClass.static_method) == "MyClass.static_method"


def test_custom_method_name_fqn_true():
    my_class = MyClass()

    os.environ["AUTODYNATRACE_CUSTOM_SERVICE_USE_FQN"] = "TRUE"
    os.environ.pop("AUTODYNATRACE_CUSTOM_SERVICE_NAME", None)
    assert custom_wrapper.generate_method_name(module_function) == "tests.test_custom.module_function"
    assert custom_wrapper.generate_method_name(my_class.class_method) == "tests.test_custom.MyClass.class_method"
    assert custom_wrapper.generate_method_name(classmethod(my_class.class_method)) == "tests.test_custom.MyClass.class_method"
    assert custom_wrapper.generate_method_name(another_function) == "tests.test_custom.another_function"

    os.environ["AUTODYNATRACE_CUSTOM_SERVICE_NAME"] = "CustomServiceName"
    assert custom_wrapper.generate_method_name(module_function) == "tests.test_custom.module_function"
    assert custom_wrapper.generate_method_name(my_class.class_method) == "tests.test_custom.MyClass.class_method"
    assert custom_wrapper.generate_method_name(classmethod(my_class.class_method)) == "tests.test_custom.MyClass.class_method"
    assert custom_wrapper.generate_method_name(another_function) == "tests.test_custom.another_function"

    if sys.version_info[0] == 2:
        assert custom_wrapper.generate_method_name(my_class.static_method) == "tests.test_custom.static_method"
        assert custom_wrapper.generate_method_name(MyClass.static_method) == "tests.test_custom.static_method"
    else:
        assert custom_wrapper.generate_method_name(my_class.static_method) == "tests.test_custom.MyClass.static_method"
        assert custom_wrapper.generate_method_name(MyClass.static_method) == "tests.test_custom.MyClass.static_method"


def test_custom_service_instrumentation():
    assert module_function() == 1
    assert decorated_method_only() == 1
    assert decorated_service_only() == 1
    assert decorated_service_and_method() == 1


def test_decorator_with_arguments():
    assert custom_wrapper.generate_service_name(module_function, "Service") == "Service"
    assert custom_wrapper.generate_method_name(module_function, "method") == "method"



import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    @wrapt.patch_function_wrapper("concurrent.futures.thread", "_WorkItem.__init__")
    def dynatrace_submit(wrapped, instance, args, kwargs):
        fn = args[1]
        module = getattr(fn, "__module__", "thread")

        if hasattr(fn, "__class__") and getattr(fn.__class__, "__name__", "unknown") != "method":
            function_name = "__call__"
            class_name = getattr(fn.__class__, "__name__", "unknown")
        else:
            function_name = getattr(fn, "__name__", "unknown")
            if hasattr(fn, "__self__") and hasattr(fn.__self__, "__class__"):
                class_name = getattr(fn.__self__.__class__, "__name__", "unknown")
            else:
                class_name = "Unknown"
        if class_name == "module":
            method = "{}.{}".format(module, function_name)
        else:
            method = "{}.{}".format(class_name, function_name)

        with sdk.trace_custom_service(method, "Threading"):
            logger.debug("Tracing upstream thread execution {}".format(method))
            link = sdk.create_in_process_link()
            setattr(instance, "__dynatrace_link", link)
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("concurrent.futures.thread", "_WorkItem.run")
    def dynatrace_run(wrapped, instance, args, kwargs):
        link = getattr(instance, "__dynatrace_link", None)
        if link is not None:
            logger.debug("Tracing downstream thread execution")
            with sdk.trace_in_process_link(link):
                return wrapped(*args, **kwargs)
        return wrapped(*args, **kwargs)



from ...log import logger
from ...sdk import sdk

import os


from bottle import hook, request, response


def instrument():
    @hook("before_request")
    def instrument_before_request():
        try:
            # extract host and port from the request
            host = request.environ.get("HTTP_HOST", "unknown")
            app_name = "{}".format(host)

            virtual_host = os.environ.get("AUTODYNATRACE_VIRTUAL_HOST", "{}".format(host))
            app_name = os.environ.get("AUTODYNATRACE_APPLICATION_ID", "Bottle ({})".format(app_name))
            context_root = os.environ.get("AUTODYNATRACE_CONTEXT_ROOT", "/")

            # Create the oneagent web app
            web_app_info = sdk.create_web_application_info(virtual_host, app_name, context_root)

            with web_app_info:
                # Attempt to extract the x-dynatrace header from the request
                dynatrace_header = request.headers.get("x-dynatrace")
                logger.debug("Bottle - tracing incoming request ({}) with header: {}".format(request.url, dynatrace_header))
                tracer = sdk.trace_incoming_web_request(web_app_info, request.url, request.method, headers=request.headers, str_tag=dynatrace_header)
                tracer.start()
                setattr(request, "__dynatrace_tracer", tracer)
        except Exception as e:
            logger.warning("Bottle - failed to instrument request: {}".format(e))

    @hook("after_request")
    def instrument_after_request():
        try:
            # check if the request was instrumented
            tracer = getattr(request, "__dynatrace_tracer", None)
            if tracer:
                tracer.set_status_code(response.status_code)
                tracer.add_response_headers(response.headers)
                logger.debug("Bottle - ending incoming request ({}) with status: {}".format(request.url, response.status_code))
                tracer.end()
        except Exception as e:
            logger.warning("Bottle - failed to instrument response: {}".format(e))



import django
import wrapt

from ...log import logger
from ...sdk import sdk

from .middlewares import insert_dynatrace_middleware
from .db import instrument_db
from .log import instrument_log


def instrument():
    @wrapt.patch_function_wrapper("django", "setup")
    def dynatrace_setup(wrapped, instance, args, kwargs):
        from django.conf import settings
        autodynatrace_wrapper_django = "autodynatrace.wrappers.django"
        if autodynatrace_wrapper_django not in settings.INSTALLED_APPS:
            if isinstance(settings.INSTALLED_APPS, tuple):

                settings.INSTALLED_APPS = settings.INSTALLED_APPS + (autodynatrace_wrapper_django,)
            else:
                settings.INSTALLED_APPS.append(autodynatrace_wrapper_django)

            logger.debug("Added autodynatrace to settings.INSTALLED_APPS")
        wrapped(*args, **kwargs)


def instrument_django():
    insert_dynatrace_middleware()
    instrument_db()
    instrument_log()



import redis
import wrapt
import oneagent
import socket

from ...log import logger
from ...sdk import sdk

from .utils import format_command_args


def instrument():
    if redis.VERSION < (3, 0, 0):
        wrapt.wrap_function_wrapper("redis", "StrictRedis.execute_command", dynatrace_execute_command)
        wrapt.wrap_function_wrapper("redis.client", "BasePipeline.execute", dynatrace_execute_pipeline)
        wrapt.wrap_function_wrapper("redis.client", "BasePipeline.immediate_execute_command", dynatrace_execute_command)
    else:
        wrapt.wrap_function_wrapper("redis", "Redis.execute_command", dynatrace_execute_command)
        wrapt.wrap_function_wrapper("redis.client", "Pipeline.execute", dynatrace_execute_command)
        wrapt.wrap_function_wrapper("redis.client", "Pipeline.immediate_execute_command", dynatrace_execute_command)


def dynatrace_execute_command(func, instance, args, kwargs):
    host = instance.connection_pool.connection_kwargs.get("host", None)
    port = instance.connection_pool.connection_kwargs.get("port", None)

    channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.OTHER, None)
    if host and port:
        try:
            socket.inet_pton(socket.AF_INET6, host)
            host = "[{}]".format(host)
        except Exception:
            pass
        channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(host, port))

    db_info = sdk.create_database_info("Cache", "Redis", channel)

    query = format_command_args(args)
    with sdk.trace_sql_database_request(db_info, "{}".format(query)):
        logger.debug("Tracing from Redis: {} {}".format(func.__name__, query))
        return func(*args, **kwargs)


def dynatrace_execute_pipeline(func, instance, args, kwargs):
    cmds = [format_command_args(c) for c, _ in instance.command_stack]
    with sdk.trace_custom_service(func.__name__, "Redis"):
        logger.debug("Tracing from Redis: {} {}".format(func.__name__, cmds))
        sdk.add_custom_request_attribute("Queries", cmds)

        return func(*args, **kwargs)



import pika
import wrapt

from ...log import logger
from ...sdk import sdk

import oneagent


def instrument():
    @wrapt.patch_function_wrapper("pika.channel", "Channel.basic_publish")
    def basic_publish_dynatrace(wrapped, instance, args, kwargs):

        try:
            host = instance.connection.params.host
            port = instance.connection.params.port
            exchange = kwargs.get("exchange")
            routing_key = kwargs.get("routing_key")

        except Exception as e:
            logger.warn("autodynatrace - could not instrument Channel.basic_publish: {}".format(e))
            return wrapped(*args, **kwargs)

        properties = kwargs.get("properties")
        if properties is not None:
            if properties.headers is None:
                properties.headers = {}
        else:
            props = pika.BasicProperties(headers={})
            kwargs["properties"] = props

        messaging_system = sdk.create_messaging_system_info(
            "RabbitMQ",
            routing_key,
            oneagent.common.MessagingDestinationType.QUEUE,
            oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(host, port)),
        )
        with messaging_system:
            with sdk.trace_outgoing_message(messaging_system) as tracer:
                tag = tracer.outgoing_dynatrace_string_tag.decode("utf-8")
                kwargs["properties"].headers[oneagent.common.DYNATRACE_MESSAGE_PROPERTY_NAME] = tag
                logger.debug(
                    "autodynatrace - Tracing RabbitMQ host={}, port={}, routing_key={}, tag={}".format(
                        host, port, routing_key, tag
                    )
                )
                return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("pika.adapters.blocking_connection", "BlockingChannel._on_consumer_message_delivery")
    def on_message_callback_dynatrace(wrapped, instance, args, kwargs):

        try:
            channel, method_frame, header_frame, body = args
            host = channel.connection.params.host
            port = channel.connection.params.port
            routing_key = method_frame.routing_key
            exchange = method_frame.exchange
            tag = None
            if header_frame is not None and header_frame.headers is not None:
                tag = header_frame.headers.get(oneagent.common.DYNATRACE_MESSAGE_PROPERTY_NAME, None)

        except Exception as e:
            logger.warn("autodynatrace - Could not trace BlockingChannel._on_consumer_message_delivery: {}".format(e))
            return wrapped(*args, **kwargs)

        messaging_system = sdk.create_messaging_system_info(
            "RabbitMQ",
            routing_key,
            oneagent.common.MessagingDestinationType.QUEUE,
            oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(host, port)),
        )
        with messaging_system:
            with sdk.trace_incoming_message_receive(messaging_system):
                logger.debug(
                    "autodynatrace - Tracing RabbitMQ host={}, port={}, routing_key={}, tag={}".format(
                        host, port, routing_key, tag
                    )
                )
                with sdk.trace_incoming_message_process(messaging_system, str_tag=tag):
                    return wrapped(*args, **kwargs)



import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    @wrapt.patch_function_wrapper("suds.client", "Client.__init__")
    def dynatrace_client(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("CreateClient", "Soap"):
            logger.debug("Tracing suds Client.__init__")
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("suds.client", "SoapClient.invoke")
    def dynatrace_client(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service(instance.method.name, "Soap"):
            logger.debug("Tracing suds Client.{}".format(instance.method.name))
            return wrapped(*args, **kwargs)



import os

from six.moves.urllib import parse

from ...log import logger

try:
    from django.core.urlresolvers import resolve
except ImportError:
    from django.urls import resolve


def get_host(request):
    host = None
    try:
        host = request.get_host()
    except Exception:
        logger.debug("Failed to get Django request host", exc_info=True)

    if not host:
        try:
            if "HTTP_HOST" in request.META:
                host = request.META["HTTP_HOST"]
            else:
                host = request.META["SERVER_NAME"]
                port = str(request.META["SERVER_PORT"])
                if port != ("443" if request.is_secure() else "80"):
                    host = "{0}:{1}".format(host, port)
        except Exception:
            logger.debug("Failed to build Django request host", exc_info=True)
            host = "unknown"

    return str(host)


def get_request_uri(request):

    host = get_host(request)
    scheme = request.scheme or ""
    return parse.urlunparse(parse.ParseResult(scheme=scheme, netloc=host, path=request.path, params="", query="", fragment=""))


def get_app_name(request):
    try:
        app_name = resolve(request.path).kwargs.get("name")
        logger.debug("Resolved app_name as '{}' from resolving the path".format(app_name))
        if app_name is None:
            app_name = "{}:{}".format(request.META.get("SERVER_NAME"), request.META.get("SERVER_PORT"))
    except Exception:
        app_name = "{}:{}".format(request.META.get("SERVER_NAME"), request.META.get("SERVER_PORT"))
        logger.debug("Could not get app name, using default: {}".format(app_name))

    return os.environ.get("AUTODYNATRACE_APPLICATION_ID", "Django ({})".format(app_name))



import functools
import os

from ...log import logger
from ...sdk import sdk


def use_fully_qualified_name():
    return (
        os.environ.get("AUTODYNATRACE_CUSTOM_SERVICE_USE_FQN")
        and os.environ.get("AUTODYNATRACE_CUSTOM_SERVICE_USE_FQN").lower() == "true"
    )


def get_custom_defined_service_name():
    return os.environ.get("AUTODYNATRACE_CUSTOM_SERVICE_NAME")


def generate_service_name(wrapped, service=None):
    if service is not None and isinstance(service, str):
        return service
    if get_custom_defined_service_name():
        return get_custom_defined_service_name()
    else:
        return get_module_path(wrapped)


def get_module_path(wrapped):
    module_path = "unknown"
    if hasattr(wrapped, "__module__"):
        module_path = wrapped.__module__
    elif hasattr(wrapped, "__func__"):
        return get_module_path(wrapped.__func__)

    class_name = None
    qual_name = None
    result = module_path

    if hasattr(wrapped, "im_class"):
        class_name = wrapped.im_class.__name__
        result = class_name

    if hasattr(wrapped, "__qualname__") and "." in wrapped.__qualname__:
        qual_name = wrapped.__qualname__.split(".")[0]
        result = qual_name

    if use_fully_qualified_name():
        result = ".".join([i for i in [module_path, class_name, qual_name] if i])

    return result


def generate_method_name(wrapped, method=None):
    if method is not None:
        return method
    name = "unknown"
    if hasattr(wrapped, "__name__"):
        name = wrapped.__name__
    elif hasattr(wrapped, "__func__"):
        return generate_method_name(wrapped.__func__)

    if get_custom_defined_service_name() or use_fully_qualified_name():
        path = get_module_path(wrapped)
        return "{}.{}".format(path, name)
    else:
        return name


def dynatrace_custom_tracer(service=None, method=None):
    def dynatrace_decorator(wrapped):
        @functools.wraps(wrapped)
        def wrapper(*args, **kwargs):
            service_name = generate_service_name(wrapped, service)
            method_name = generate_method_name(wrapped, method)

            with sdk.trace_custom_service(method_name, service_name):
                logger.debug("Custom tracing - {}: {}".format(service_name, method_name))
                return wrapped(*args, **kwargs)

        return wrapper

    if callable(service):
        return dynatrace_decorator(service)
    else:
        return dynatrace_decorator



import os
import logging
from ...sdk import sdk

auto_log_correlation = os.environ.get("AUTODYNATRACE_AUTOMATIC_LOG_CORRELATION", "False").strip().lower() in (
"true", "1")

def instrument_log():
    old_factory = logging.getLogRecordFactory()

    def record_factory(*args, **kwargs):
        record = old_factory(*args, **kwargs)
        if sdk:
            trace_info = sdk.tracecontext_get_current()
            span_id, trace_id = trace_info.span_id, trace_info.trace_id
            record.span_id = span_id
            record.trace_id = trace_id

        return record

    if auto_log_correlation:
        logging.setLogRecordFactory(record_factory)


import six
import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    httplib = six.moves.http_client

    def dynatrace_putrequest(wrapped, instance, args, kwargs):
        method, path = args[:2]
        scheme = "https" if "HTTPS" in instance.__class__.__name__ else "http"
        url = "{}://{}{}{}".format(
            scheme, instance.host, ":{}".format(instance.port) if str(instance.port) not in ["80", "443"] else "", path
        )
        tracer = sdk.trace_outgoing_web_request(url, method)
        tracer.start()
        setattr(instance, "__dynatrace_tracer", tracer)
        ret = wrapped(*args, **kwargs)
        tag = tracer.outgoing_dynatrace_string_tag
        logger.debug("Tracing urllib, url: '{}', tag: '{}'".format(url, tag))
        instance.putheader("x-dynatrace", tag)
        return ret

    def dynatrace_getresponse(wrapped, instance, args, kwargs):
        tracer = getattr(instance, "__dynatrace_tracer", None)
        response = wrapped(*args, **kwargs)
        # print(traceback.print_stack())

        if tracer is not None:
            tracer.set_status_code(response.status)
            tracer.end()
            delattr(instance, "__dynatrace_tracer")

        return response

    setattr(httplib.HTTPConnection, "putrequest", wrapt.FunctionWrapper(httplib.HTTPConnection.putrequest, dynatrace_putrequest))
    setattr(
        httplib.HTTPConnection, "getresponse", wrapt.FunctionWrapper(httplib.HTTPConnection.getresponse, dynatrace_getresponse)
    )



from setuptools import setup, find_packages

setup(
    name="jsm-autodynatrace",
    version="1.1.1",
    packages=find_packages(),
    package_data={"autodynatrace": ["wrappers/*"]},
    install_requires=["wrapt>=1.11.2", "oneagent-sdk>=1.3.0", "six>=1.10.0", "autowrapt>=1.0"],
    tests_require=["pytest", "mock", "tox", "django"],
    entry_points={"autodynatrace": ["string = autodynatrace:load"]},
    python_requires=">=3.6",
    author="Juntos Somos Mais",
    author_email="labs@juntossomosmais.com.br",
    description="Auto instrumentation for the OneAgent SDK",
    long_description="The autodynatrace package will auto instrument your python apps",
    url="https://github.com/juntossomosmais/OneAgent-SDK-Python-AutoInstrumentation",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved",
        "License :: OSI Approved :: Apache Software License",  # 2.0
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Topic :: System :: Monitoring",
    ],
    project_urls={"Issue Tracker": "https://github.com/juntossomosmais/OneAgent-SDK-Python-AutoInstrumentation/issues"},
)



from django.apps import AppConfig, apps
from ...log import logger
from .wrapper import instrument_django


class DynatraceConfig(AppConfig):
    name = "autodynatrace.wrappers.django"
    label = "dynatrace_django"

    def ready(self):
        instrument_django()



import os
import wrapt

from ...log import logger
from ...sdk import sdk


def instrument():
    @wrapt.patch_function_wrapper("tornado.web", "RequestHandler._execute")
    def patch_execute(wrapped, instance, args, kwargs):
        request = instance.request

        virtual_host = os.environ.get("AUTODYNATRACE_VIRTUAL_HOST", request.host)
        app_name = os.environ.get("AUTODYNATRACE_APPLICATION_ID", "Tornado")
        context_root = os.environ.get("AUTODYNATRACE_CONTEXT_ROOT", "/")
        wappinfo = sdk.create_web_application_info(virtual_host, app_name, context_root)

        dt_headers = None
        dt_header = instance.request.headers.get("X-Dynatrace")
        if os.environ.get("AUTODYNATRACE_CAPTURE_HEADERS", False):
            dt_headers = dict(instance.request.headers)

        with wappinfo:
            url = instance.request.full_url().rsplit("?", 1)[0]
            logger.debug("tornado - tracing '{}' with header: '{}'".format(url, dt_header))
            method = instance.request.method
            tracer = sdk.trace_incoming_web_request(wappinfo, url, method, headers=dt_headers, str_tag=dt_header)
            tracer.start()
            sdk.add_custom_request_attribute("query", instance.request.query)

            setattr(instance.request, "__dynatrace_tracer", tracer)
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("tornado.web", "RequestHandler.on_finish")
    def patch_on_finish(wrapped, instance, args, kwargs):
        tracer = getattr(instance.request, "__dynatrace_tracer", None)
        if tracer:
            tracer.set_status_code(instance.get_status())
            tracer.end()

        return wrapped(*args, **kwargs)



import re
import socket

import cx_Oracle
import wrapt

from ...log import logger
from ...sdk import sdk


import oneagent


class DynatraceConnection(cx_Oracle.Connection):
    def __init__(self, *args, **kwargs):
        super(DynatraceConnection, self).__init__(*args, **kwargs)


class DynatraceCursor(cx_Oracle.Cursor):
    def __init__(self, *args, **kwargs):
        super(DynatraceCursor, self).__init__(*args, **kwargs)


def instrument():
    def trace_query(wrapped, instance, args, kwargs):
        if args:
            query = args[0]
            dsn = instance.connection.dsn or "unknown_oracle"
            host = None
            port = 1521

            host_match = re.search(r"HOST=(.*?)\)", dsn)
            if host_match and host_match.groups():
                host = host_match.group(1)

            port_match = re.search(r"PORT=(.*?)\)", dsn)
            if port_match and port_match.groups():
                port = port_match.group(1)

            service_name = dsn
            service_match = re.search(r"SERVICE_NAME=(.*?)\)", dsn)
            if service_match and service_match.groups():
                service_name = service_match.group(1)

            channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.OTHER, "Oracle")
            if host is not None:
                try:
                    socket.inet_pton(socket.AF_INET6, host)
                    host = "[{}]".format(host)
                except Exception:
                    pass
                channel = oneagent.sdk.Channel(oneagent.sdk.ChannelType.TCP_IP, "{}:{}".format(host, port))

            db_info = sdk.create_database_info(service_name, oneagent.sdk.DatabaseVendor.ORACLE, channel)
            with sdk.trace_sql_database_request(db_info, "{}".format(query)):
                logger.debug(
                    "Tracing cx_Oracle query: '{}', host: '{}', port: '{}', database: '{}'".format(
                        query, host, port, service_name
                    )
                )
                return wrapped(*args, **kwargs)

        return wrapped(*args, **kwargs)

    def fetchone_wrapper(wrapped, instance, args, kwargs):
        if not getattr(instance, "_dt_fetchone_reported", False):
            with sdk.trace_custom_service("fetchone", "cx_Oracle"):
                setattr(instance, "_dt_fetchone_reported", True)
                sdk.add_custom_request_attribute("Note", "Only the first fetch (slowest) is recorded")
                return wrapped(*args, **kwargs)
        return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("cx_Oracle", "connect")
    def connect_dynatrace(wrapped, instance, args, kwargs):
        return DynatraceConnection(*args, **kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.cx_Oracle.wrapper", "DynatraceConnection.cursor")
    def cursor_dynatrace(wrapped, instance, args, kwargs):
        return DynatraceCursor(instance, *args, **kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.cx_Oracle.wrapper", "DynatraceCursor.execute")
    def execute_dynatrace(wrapped, instance, args, kwargs):
        return trace_query(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.cx_Oracle.wrapper", "DynatraceCursor.executemany")
    def execute_many_dynatrace(wrapped, instance, args, kwargs):
        return trace_query(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.cx_Oracle.wrapper", "DynatraceCursor.fetchone")
    def fetchone_dynatrace(wrapped, instance, args, kwargs):
        return fetchone_wrapper(wrapped, instance, args, kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.cx_Oracle.wrapper", "DynatraceCursor.fetchmany")
    def fetchmany_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("fetchmany", "cx_Oracle"):
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.cx_Oracle.wrapper", "DynatraceCursor.fetchall")
    def fetchall_dynatrace(wrapped, instance, args, kwargs):
        with sdk.trace_custom_service("fetchall", "cx_Oracle"):
            return wrapped(*args, **kwargs)

    @wrapt.patch_function_wrapper("autodynatrace.wrappers.cx_Oracle.wrapper", "DynatraceCursor.__next__")
    def next_dynatrace(wrapped, instance, args, kwargs):
        return fetchone_wrapper(wrapped, instance, args, kwargs)



import wrapt

from ...log import logger
from ...sdk import sdk
from ..utils import normalize_vendor


class TracedCursor(wrapt.ObjectProxy):
    def __init__(self, cursor, db_info):
        super(TracedCursor, self).__init__(cursor)
        self.db_info = db_info
        self._self_last_execute_operation = None
        self._original_cursor = cursor

    def _trace_method(self, method, query, *args, **kwargs):

        # It could be psycopg2.sql.Composable, but we don't want to import that here
        if not isinstance(query, str):
            try:
                query = query.as_string(self._original_cursor)
            except Exception:
                pass
        logger.debug("Tracing Database Call '{}' to {}".format(str(query), self.db_info))

        try:
            with sdk.trace_sql_database_request(self.db_info, f"{query}"):
                return method(*args, **kwargs)
        except Exception as e:
            logger.warning(f"Error instrumenting database: {e}")
            return method(*args, **kwargs)

    def executemany(self, query, *args, **kwargs):
        self._self_last_execute_operation = query
        return self._trace_method(self.__wrapped__.executemany, query, query, *args, **kwargs)

    def execute(self, query, *args, **kwargs):
        self._self_last_execute_operation = query
        return self._trace_method(self.__wrapped__.execute, query, query, *args, **kwargs)

    def callproc(self, proc, args):
        self._self_last_execute_operation = proc
        return self._trace_method(self.__wrapped__.callproc, proc, proc, args)

    def __enter__(self):
        self.__wrapped__.__enter__()
        return self


class TracedConnection(wrapt.ObjectProxy):
    def __init__(self, conn, cursor_cls=None):

        if not cursor_cls:
            cursor_cls = TracedCursor
        super(TracedConnection, self).__init__(conn)
        self._self_cursor_cls = cursor_cls

    def _trace_method(self, method, *args, **kwargs):
        logger.info("Tracing Connection {}".format(args))
        return method(*args, **kwargs)

    def cursor(self, *args, **kwargs):
        return self.__wrapped__.cursor(*args, **kwargs)

    def commit(self, *args, **kwargs):
        span_name = "commit"
        return self._trace_method(self.__wrapped__.commit, span_name, {}, *args, **kwargs)

    def rollback(self, *args, **kwargs):
        span_name = "rollback"
        return self._trace_method(self.__wrapped__.rollback, span_name, {}, *args, **kwargs)


def _get_vendor(conn):
    try:
        name = _get_module_name(conn)
    except Exception:
        logger.debug("couldnt parse module name", exc_info=True)
        name = "sql"
    return normalize_vendor(name)


def _get_module_name(conn):
    return conn.__class__.__module__.split(".")[0]



import flask
from wsgiref.util import request_uri
import wrapt

from ...log import logger
from ...sdk import sdk
import socket
import os


def instrument():
    @wrapt.patch_function_wrapper("flask", "Flask.full_dispatch_request")
    def full_dispatch_request_dynatrace(wrapped, instance, args, kwargs):
        try:
            env = flask.request.environ
            method = env.get("REQUEST_METHOD", "GET")
            url = env.get("REQUEST_URI") or env.get("RAW_URI") or env.get("werkzeug.request").url or request_uri(env)
            host = env.get("SERVER_NAME") or socket.gethostname() or "localhost"
            app_name = flask.current_app.name

            dt_headers = None
            dt_header = flask.request.headers.get("X-Dynatrace")
            if os.environ.get("AUTODYNATRACE_CAPTURE_HEADERS", False):
                dt_headers = dict(flask.request.headers)

            virtual_host = os.environ.get("AUTODYNATRACE_VIRTUAL_HOST", "{}".format(host))
            app_name = os.environ.get("AUTODYNATRACE_APPLICATION_ID", "Flask ({})".format(app_name))
            context_root = os.environ.get("AUTODYNATRACE_CONTEXT_ROOT", "/")

            wappinfo = sdk.create_web_application_info(virtual_host, app_name, context_root)

        except Exception as e:
            logger.debug("dynatrace - could not instrument: {}".format(e))
            return wrapped(*args, **kwargs)

        with wappinfo:
            logger.debug("Tracing with header: {}".format(dt_header))
            tracer = sdk.trace_incoming_web_request(wappinfo, url, method, headers=dt_headers, str_tag=dt_header)
            tracer.start()
            logger.debug("dynatrace - full_dispatch_request_dynatrace: {}".format(url))
            setattr(flask.request, "__dynatrace_tracer", tracer)
            response = wrapped(*args, **kwargs)
            tracer.set_status_code(response.status_code)
            tracer.end()
            return response

    @wrapt.patch_function_wrapper("flask", "Flask.handle_exception")
    def handle_exception_dynatrace(wrapped, instance, args, kwargs):
        tracer = getattr(flask.request, "__dynatrace_tracer", None)
        if tracer is not None:
            exception_type = type(args[0]).__name__
            exception_message = str(args[0])

            logger.debug("Reporting flask exception: {} - {}".format(exception_type, exception_message))
            tracer.set_status_code(500)
            tracer.mark_failed(exception_type, exception_message)
            tracer.end()
        return wrapped(*args, **kwargs)



import sys

from celery import signals
from celery import registry

from ...log import logger
from ...sdk import sdk, shutdown, init
import oneagent


DT_KEY = "__dynatrace_tag"

CELERY_WORKER_PROCESS = sys.argv and sys.argv[0].endswith("celery") and "worker" in sys.argv

if CELERY_WORKER_PROCESS:
    logger.debug("It looks like we are initializing Celery workers, reinitializing with forkable=True")
    shutdown()
    init(forkable=True)


def instrument():
    def get_task_id(context):
        headers = context.get("headers")
        body = context.get("body")
        if headers:
            return headers.get("id")
        else:
            return body.get("id")

    def add_tracer(task, task_id, tracer):
        tracer_dict = getattr(task, DT_KEY, None)
        if tracer_dict is None:
            tracer_dict = dict()
            setattr(task, DT_KEY, tracer_dict)

        tracer_dict[task_id] = tracer

    def get_tracer(task, task_id):
        tracer_dict = getattr(task, DT_KEY, None)
        if tracer_dict is not None:
            return tracer_dict.get(task_id, None)

    def remove_tracer(task, task_id):
        tracer_dict = getattr(task, DT_KEY, None)
        if tracer_dict is not None:
            tracer_dict.pop(task_id, None)

    def dt_after_task_publish(**kwargs):
        # This is executed on the app side, after a task has been sent to Celery
        task_name = kwargs.get("sender")
        task = registry.tasks.get(task_name)
        task_id = get_task_id(kwargs)

        if task is None or task_id is None:
            logger.debug("Could not obtain task or task_id")
            return

        tracer = get_tracer(task, task_id)
        if tracer is not None:
            tracer.end()
            remove_tracer(task, task_id)

    def dt_before_task_publish(**kwargs):
        # This is executed on the app side, before a task has been sent to Celery
        task_name = kwargs.get("sender")
        task = registry.tasks.get(task_name)
        task_id = get_task_id(kwargs)

        if task is None or task_id is None:
            logger.debug("Could not obtain task or task_id")
            return

        msi_handle = sdk.create_messaging_system_info(
            "Celery",
            kwargs.get("routing_key", "celery"),
            oneagent.common.MessagingDestinationType.QUEUE,
            oneagent.sdk.Channel(oneagent.sdk.ChannelType.OTHER, "celery"),
        )

        with msi_handle:
            tracer = sdk.trace_outgoing_message(msi_handle)
            tracer.start()
            tag: str = tracer.outgoing_dynatrace_string_tag.decode("utf-8")
            logger.debug("Celery - inserting tag {}".format(tag))
            kwargs["headers"][DT_KEY] = tag
            add_tracer(task, task_id, tracer)

    def dt_task_prerun(*args, **kwargs):
        # This is executed on the worker, before a task is run
        task = kwargs.get("sender")
        task_id = kwargs.get("task_id")
        if task is None or task_id is None:
            logger.debug("Could not obtain task or task_id")
            return

        msi_handle = sdk.create_messaging_system_info(
            "Celery",
            task.request.delivery_info.get("routing_key", "celery"),
            oneagent.common.MessagingDestinationType.QUEUE,
            oneagent.sdk.Channel(oneagent.sdk.ChannelType.OTHER),
        )

        with msi_handle:
            with sdk.trace_incoming_message_receive(msi_handle):
                tag = getattr(task.request, DT_KEY, "")
                logger.debug("Celery - received tag: {}".format(tag))
                tracer = sdk.trace_incoming_message_process(msi_handle, str_tag=tag)
                tracer.start()
                add_tracer(task, task_id, tracer)

    def dt_task_postrun(*args, **kwargs):
        # This is executed on the worker, after a task is run
        task = kwargs.get("sender")
        task_id = kwargs.get("task_id")

        if task is None or task_id is None:
            logger.debug("Could not obtain task or task_id")
            return

        tracer = get_tracer(task, task_id)
        if tracer is not None:
            tracer.end()
            remove_tracer(task, task_id)

    signals.task_prerun.connect(dt_task_prerun, weak=False)
    signals.task_postrun.connect(dt_task_postrun, weak=False)
    signals.after_task_publish.connect(dt_after_task_publish, weak=False)
    signals.before_task_publish.connect(dt_before_task_publish, weak=False)



import oneagent
import atexit

from .log import logger

sdk = None


def shutdown():
    oneagent.shutdown()


def init(forkable=False):
    global sdk
    oneagent.initialize(forkable=forkable)
    state = oneagent.get_sdk().agent_state
    logger.debug("Initialized autodynatrace with AgentState: {}".format(state))
    if state != oneagent.common.AgentState.ACTIVE:
        logger.warning("Could not initialize the OneAgent SDK, AgentState: {}".format(state))

    atexit.register(shutdown)
    sdk = oneagent.get_sdk()
